/* Copyright (C) 1992 by R.Harmsen
   Interface to FM-synthesizer chip, AdLib or Soundblaster.
 */

#ifndef FMCHIP_H
#define FMCHIP_H

/* 27 September 2020: Added and used this typedef */
typedef unsigned short int fmchip_value_t;

/* All function-parameters may be PREV_VAL, meaning "same as in last call" */
/* @? not fully implemented */
#define PREV_VAL -1

/* The chip has 9 channels (1 thru 9), which may be used as a
   carrier-modulator pair each (synth_mode = FREQ_MODU_SYNTH).
   Modul_carr then specifies which part of a channel is meant.
   If synth_mode is ADDITIVE_SYNTH however, each channel has two
   more or less independent sine-generators.

   All the functions that take a channel number and a modcar argument,
   have an alternative macro-form, which takes a corresponding voice
   number 0 through 17 instead. These macros have "setv" in the name where
   the functions have "set".
*/

/* Frequencies, precalculated to integer values for efficiently
   setting the Sound Blaster's frequency. Indices in the table
   correspond to MIDI-notes, where number 69 is the 440 Hz a.
   Because the table even supports quarter tones (24-tones per octave),
   you should shift a MIDI-tone number one bit to the left to get
   normal tones.

   See also the two prototypes for set_frequency, and freq2value.
   This array is initialized with proper values in fmchip.c.
   Code to generate the initialisation code is there too, but
   made inactive. See in the bottom of this source file.
 */
#define SBFREQ24_ELEM ((unsigned int)(109 * 24.0 / 12.0 + 0.5))
#define SBFREQ53_ELEM ((unsigned int)(109 * 53.0 / 12.0 + 0.5))
extern fmchip_value_t sbfreq24_value[SBFREQ24_ELEM];
extern fmchip_value_t sbfreq53_value[SBFREQ53_ELEM];

enum modul_carr
{
   MODULATOR,
   CARRIER,
};

enum synth_mode
{
   ADDITIVE_SYNTH = 0,
   FREQMODU_SYNTH = 1,
};


/**************************************************************************
   Returns 0 if no chip found, 1 if monophonic AdLib or Soundblaster
   detected, or 2 if stereophonic Soundblaster detected.
 *************************************************************************/
int init_fmchip (void);


/**************************************************************************

 *************************************************************************/
void set_envelope (
   int channel,
   enum modul_carr modcar,
   int sustain,
   int attack,
   int decay,
   int sustain_lvl,
   int release);

#define setv_envelope(voice, e, a, d, s, r) \
   set_envelope((voice) / 2 + 1, (voice) % 2 ? CARRIER : MODULATOR, \
      (e), (a), (d), (s), (r))

/**************************************************************************

 *************************************************************************/
void set_frequency_f (int channel, double freq);

/**************************************************************************
   Less intuitive, but much faster interface. A function to pre-calculate
   the value is available: See freq2value below.
 *************************************************************************/
void set_frequency_v (int channel, fmchip_value_t value);

/**************************************************************************
   Conversion between frequency in hertz, and internal soundblaster value,
   concisting of 10 bits, plus a 3 bits octave counter.

   Two functions, for both directions.
 *************************************************************************/
fmchip_value_t freq2value (double freq);
double value2freq (fmchip_value_t value);

/**************************************************************************

 *************************************************************************/
void set_multiplier (int channel, enum modul_carr modcar, int multiplier);

#define setv_multiplier(voice, m) \
   set_multiplier((voice) / 2 + 1, ((voice) & 0x1) ? CARRIER : MODULATOR, \
   (m))


/**************************************************************************
   0 is the maximum level, minimum is 63, which gives 48dB attenuation
 *************************************************************************/
void set_level (int channel, enum modul_carr modcar, int level);

#define setv_level(voice, l) \
   set_level((voice) / 2 + 1, ((voice) & 0x1) ? CARRIER : MODULATOR, (l))


/**************************************************************************

 *************************************************************************/
void set_synth_mode  (int channel, enum synth_mode synth);
void set_fm_feedback (int channel, int feedback);


/**************************************************************************
  Key scale level: ksl may be between 0 minimum, and 3 maximum
 *************************************************************************/
void set_ksl (int channel, enum modul_carr modcar, int ksl);

#define setv_ksl(voice, k) \
   set_ksl((voice) / 2 + 1, ((voice) & 0x1) ? CARRIER : MODULATOR, (k))

/**************************************************************************
  Key scale rate: ksr may be 0 minimum, or 1 maximum
 *************************************************************************/
void set_ksr (int channel, enum modul_carr modcar, int ksr);

#define setv_ksr(voice, k) \
   set_ksr((voice) / 2 + 1, ((voice) & 0x1) ? CARRIER : MODULATOR, (k))


/**************************************************************************
  Am and vib may be 0 or 1, to indicate amplitude vibrato and/or
  frequency vibrato no/yes.
 *************************************************************************/
void set_am_vib (int channel, enum modul_carr modcar, int am, int vib);

#define setv_am_vib(voice, a, v) \
   set_am_vib((voice) / 2 + 1, ((voice) & 0x1) ? CARRIER : MODULATOR, \
   (a), (v))

/**************************************************************************
  Am and vib may be 0 or 1, to indicate amplitude vibrato and/or
  frequency vibrato intensity low/high.
 *************************************************************************/
void seti_am_vib (int am, int vib);


/**************************************************************************
  Waveform setting WS: 0 to 3
 *************************************************************************/
void set_ws (int channel, enum modul_carr modcar, int ws);

#define setv_ws(voice, w) \
   set_ksl((voice) / 2 + 1, ((voice) & 0x1) ? CARRIER : MODULATOR, (w))


/**************************************************************************
  Key scale level: ksl may be between 0 minimum, and 3 maximum
 *************************************************************************/
void set_ksl (int channel, enum modul_carr modcar, int ksl);

#define setv_ksl(voice, k) \
   set_ksl((voice) / 2 + 1, ((voice) & 0x1) ? CARRIER : MODULATOR, (k))


/**************************************************************************

 *************************************************************************/
void set_percussion (int yes_no);


/**************************************************************************

 *************************************************************************/
void note_on (int channel);


/**************************************************************************

 *************************************************************************/
void note_off (int channel);


/* Values for perc_instrument in perc_on */
#define K_BASS_DRUM  0x10
#define K_SNARE_DRUM 0x08
#define K_TOM_TOM    0x04
#define K_TOP_CYMBAL 0x02
#define K_HI_HAT     0x01

/* Channel values for setting the frequency of percussion instruments */
#define F_BASS_DRUM  7
#define F_TOM_TOM    9

/* Values for perc_instrument envelope settings */
/* For the bass drum, the carrier/modulator must be added, for the
   others this is included in the macro */
#define E_BASS_DRUM  7
#define E_SNARE_DRUM 8, CARRIER
#define E_TOM_TOM    9, MODULATOR
#define E_TOP_CYMBAL 9, CARRIER
#define E_HI_HAT     8, MODULATOR

/**************************************************************************

 *************************************************************************/
void perc_on (int perc_instrument);

/**************************************************************************
  Set left, right or both channels. If the Soundblaster or Adlib does
  not have stereo capability, any calls works out as both.
  (NOT yet IMPLEMENTED, does not work, call ignored)
 *************************************************************************/
enum left_right {SB_LEFT = -1, SB_BOTH = 0, SB_RIGHT = 1};
void set_left_right (enum left_right lrb);

/**************************************************************************
  15 October 2020, 01:30B, I woke up and suddenly understood what this bit
  (bit 6 of register 0x08) is for: when set, the 10 bits value for frequency
  control get an extra 11th bits, implicit and unstored, because it always
  have the value of 1. This improved the resulution, because the useful
  range of 511 to 1023 is replaced by 1024 to 2047.
  Compare the hidden bit (implicit bit) of the significand (mantisse) of
  IEEE 32- and 64-bits floating point formats. Same principle.

  PS, during the day, 15 October: I thought it would make sense if it worked
  that way, but it doesn't. Testing with DosBox 0.74, in OPL2 and OPL3 mode,
  I noticed no effect at all.

  17 October 2020: Now I got it. See http://www.opensound.com/dmguide/dmfm.html
  and other sites with the same text, where it says:

  "kbd_scale is a flag (1-bit) that turns on the keyboard scale rate. If
   kbd_scale is 1 then the attack/decay rates become faster as the pitch
   (fnum + octave) increases."
  "kbd_split This is a flag (1 bit) field which determines the split method to
   select the key scale number. This field is used to select the kbd_scale
   value. Depending on the pitch of the note (octave + fnum), a key scale
   number between 0 and 15 is generated. This key scale number is applied to
   the attack/decay/sustain/release rates depending whether kbd_scale is
   1 or 0. If kbd_split=1 then the key scale number depends on the Most
   Significant Bit (MSB) of the frequency. If kbd_split=0 then the key_scale
   number depends on the the 2nd MSB of the frequency."

  The "key scale number" is the setting KSR.
  So the background is that when bit 6 of register 8 is set, it is not not
  bit 9 of the 10-bit frequency value that determines the faster attach/decay
  for higher notes, but bit 8, because in practice, bit 9 is always 1, for
  optimal frequency accuracy. Values 511-1023 are used, never below.
 *************************************************************************/
void set_nts (int yes_no);


#endif
